/*

Copyright (C) 2002 J.D. Cole

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
  $Header: /home/jd/hpo/cvs-root/code/dld-projects/mpi/mpi_send.cc,v 1.6 2002/12/11 19:40:06 jd Exp $
*/

#include <octave_mpi.h>

#include <mpi.h>
#include <octave/oct.h>
#include <octave/byte-swap.h>

#include <sstream>

#include <stdio.h>

static bool
mpi_save_binary_data (std::ostream& os, const octave_value& tc,
		      const std::string& name, const std::string& doc,
		      bool mark_as_global, bool save_as_floats);

DEFUN_DLD(mpi_send, args, , "MPI_Send")
{
  if(!octave_mpi_verify_session())
    {
      error("mpi_send: mpi_init has not been called or"
	    " has concluded the MPI session");

      return octave_value();
    }

  if (args.length() < 2)
    {
      error("mpi_send: requires 2 arguments.");
      
      return octave_value();
    }

  // Make sure a valid, that means that it is a real vector of some size
  // and that each of the values is less than the MPI_COMM_WORD size
  int reciever_rank;

  // Should verify the rank is valid
  if (args(0).is_real_scalar())
    reciever_rank = args(0).int_value();
  else
    {
      error("mpi_send: bad reciever");

      return octave_value();
    }

  std::basic_ostringstream<char> send_buffer;

  // Save the byte format to the stream
  oct_mach_info::float_format flt_fmt = 
    oct_mach_info::native_float_format();

  send_buffer.write(X_CAST(char *,&flt_fmt),
		    sizeof(oct_mach_info::float_format));
		    

  // Serialize all each entry in binary form
  for (int i = 1; i< args.length() ; i++)
    {
      // Make sure the value exists
      if (args(i).is_defined())
	{
	  // Convert octave_value to binary
	  mpi_save_binary_data(send_buffer, 
			       args(i), "V", "", 
			       false, false);
	}
      else
	{
	  error("mpi_send: Doesn't handle that form of data yet");

	  return octave_value();
	}
    }
  
  send_buffer.eof();

  std::basic_string<char> send_data = send_buffer.str();
  int buffer_length = send_data.length();
  const char *buffer_data = send_data.data();


  // Send the length of the binary data
  MPI_Send((void*)&buffer_length, 1, MPI_INT,
	   reciever_rank, 
	   OCTAVE_MPI_TAG_SETUP, 
	   MPI_COMM_WORLD);
  
  // Actually send the data
  MPI_Send((void*)buffer_data, buffer_length, MPI_BYTE,
	   reciever_rank,
	   OCTAVE_MPI_TAG_DATA, 
	   MPI_COMM_WORLD);

  return args;
}

/* 
----------------------------------------------------------------------------
   The code below this statement (specifically functions "get_save_type" and 
   "save_binary_data" from the file load-save.cc) has been taken from the 
   GNU/Octave source code and is covered by the following license. Please 
   honor it apropriately.

   Also thanks to Andy Jacobson for his work initially on octave-mpi which uses
   the Octave load/save commands to convert variables to binary, on which, this
   effort is based. (See http://corto.icg.to.infn.it/andy/octave-mpi/)
----------------------------------------------------------------------------

Copyright (C) 1996, 1997 John W. Eaton

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

static save_type
get_save_type (double max_val, double min_val)
{
  save_type st = LS_DOUBLE;

  if (max_val < 256 && min_val > -1)
    st = LS_U_CHAR;
  else if (max_val < 65536 && min_val > -1)
    st = LS_U_SHORT;
  else if (max_val < 4294967295UL && min_val > -1)
    st = LS_U_INT;
  else if (max_val < 128 && min_val >= -128)
    st = LS_CHAR;
  else if (max_val < 32768 && min_val >= -32768)
    st = LS_SHORT;
  else if (max_val <= 2147483647L && min_val >= -2147483647L)
    st = LS_INT;

  return st;
}

// Save the data from TC along with the corresponding NAME, help
// string DOC, and global flag MARK_AS_GLOBAL on stream OS in the
// binary format described above for read_binary_data.

static bool
mpi_save_binary_data (std::ostream& os, const octave_value& tc,
		      const std::string& name, const std::string& doc,
		      bool mark_as_global, bool save_as_floats) 
{

  FOUR_BYTE_INT name_len = name.length ();

  os.write (X_CAST (char *, &name_len), 4);
  os << name;

  FOUR_BYTE_INT doc_len = doc.length ();

  os.write (X_CAST (char *, &doc_len), 4);
  os << doc;

  char tmp;

  tmp = mark_as_global;
  os.write (X_CAST (char *, &tmp), 1);

  if (tc.is_string ())
    {
      tmp = 7;
      os.write (X_CAST (char *, &tmp), 1);
      FOUR_BYTE_INT nr = tc.rows ();
      os.write (X_CAST (char *, &nr), 4);
      charMatrix chm = tc.char_matrix_value ();
      for (int i = 0; i < nr; i++)
	{
	  FOUR_BYTE_INT len = chm.cols ();
	  os.write (X_CAST (char *, &len), 4);
	  std::string tstr = chm.row_as_string (i);
	  const char *tmp = tstr.data ();
	  os.write (X_CAST (char *, tmp), len);
	}
    }
  else if (tc.is_range ())
    {
      tmp = 6;
      os.write (X_CAST (char *, &tmp), 1);
      tmp = (char) LS_DOUBLE;
      os.write (X_CAST (char *, &tmp), 1);
      Range r = tc.range_value ();
      double bas = r.base ();
      double lim = r.limit ();
      double inc = r.inc ();
      os.write (X_CAST (char *, &bas), 8);
      os.write (X_CAST (char *, &lim), 8);
      os.write (X_CAST (char *, &inc), 8);
    }
  else if (tc.is_real_scalar ())
    {
      tmp = 1;
      os.write (X_CAST (char *, &tmp), 1);
      tmp = (char) LS_DOUBLE;
      os.write (X_CAST (char *, &tmp), 1);
      double tmp = tc.double_value ();
      os.write (X_CAST (char *, &tmp), 8);
    }
  else if (tc.is_real_matrix ())
    {
      tmp = 2;
      os.write (X_CAST (char *, &tmp), 1);
      Matrix m = tc.matrix_value ();
      FOUR_BYTE_INT nr = m.rows ();
      FOUR_BYTE_INT nc = m.columns ();
      os.write (X_CAST (char *, &nr), 4);
      os.write (X_CAST (char *, &nc), 4);
      int len = nr * nc;
      save_type st = LS_DOUBLE;
      if (save_as_floats)
	{
	  if (m.too_large_for_float ())
	    {
	      warning ("save: some values too large to save as floats --");
	      warning ("save: saving as doubles instead");
	    }
	  else
	    st = LS_FLOAT;
	}
      else if (len > 8192) // XXX FIXME XXX -- make this configurable.
	{
	  double max_val, min_val;
	  if (m.all_integers (max_val, min_val))
	    st = get_save_type (max_val, min_val);
	}
      const double *mtmp = m.data ();

      write_doubles (os, mtmp, st, len);
    }
  else if (tc.is_complex_scalar ())
    {
      tmp = 3;
      os.write (X_CAST (char *, &tmp), 1);
      tmp = (char) LS_DOUBLE;
      os.write (X_CAST (char *, &tmp), 1);
      Complex tmp = tc.complex_value ();
      os.write (X_CAST (char *, &tmp), 16);
    }
  else if (tc.is_complex_matrix ())
    {
      tmp = 4;
      os.write (X_CAST (char *, &tmp), 1);
      ComplexMatrix m = tc.complex_matrix_value ();
      FOUR_BYTE_INT nr = m.rows ();
      FOUR_BYTE_INT nc = m.columns ();
      os.write (X_CAST (char *, &nr), 4);
      os.write (X_CAST (char *, &nc), 4);
      int len = nr * nc;
      save_type st = LS_DOUBLE;
      if (save_as_floats)
	{
	  if (m.too_large_for_float ())
	    {
	      warning ("save: some values too large to save as floats --");
	      warning ("save: saving as doubles instead");
	    }
	  else
	    st = LS_FLOAT;
	}
      else if (len > 4096) // XXX FIXME XXX -- make this configurable.
	{
	  double max_val, min_val;
	  if (m.all_integers (max_val, min_val))
	    st = get_save_type (max_val, min_val);
	}
      const Complex *mtmp = m.data ();
      write_doubles (os, X_CAST (const double *, mtmp), st, 2*len);
    }
  else
    gripe_wrong_type_arg ("save", tc, false);

  return os;
}
